# SV comparison pipeline
import logging
import os
import shutil
import sys
import tempfile
from os.path import basename, dirname
from os.path import join as pjoin

import pandas as pd
from simppl.simple_pipeline import SimplePipeline
from ugbio_core.exec_utils import print_and_execute
from ugbio_core.logger import logger
from ugbio_core.vcf_utils import VcfUtils
from ugbio_core.vcfbed import vcftools


class SVComparison:
    """
    SV comparison pipeline
    """

    def __init__(self, simple_pipeline: SimplePipeline | None = None, logger: logging.Logger | None = None):
        """Combines VCF in parts from GATK and indices the result

        Parameters
        ----------
        simple_pipeline : SimplePipeline, optional
            Optional SimplePipeline object for executing shell commands
        """
        self.sp = simple_pipeline
        self.vu = VcfUtils(self.sp, logger=logger)
        if logger is None:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger

    def __execute(self, command: str, output_file: str | None = None):
        """Summary

        Parameters
        ----------
        command : str
            Description
        output_file : str, optional
            Description
        """
        print_and_execute(command, output_file=output_file, simple_pipeline=self.sp, module_name=__name__)

    def run_truvari(  # noqa: PLR0913
        self,
        calls: str,
        gt: str,
        outdir: str,
        bed: str | None = None,
        pctseq: float = 0.0,
        pctsize: float = 0.0,
        maxsize: int = 50000,
        *,
        erase_outdir: bool = True,
        ignore_filter: bool = False,
        ignore_type: bool = True,
        skip_collapse: bool = False,
    ):
        """
        Run truvari, generate truvari report and concordance VCF
        Parameters
        ----------
        calls : str
            Calls file
        gt : str
            Ground truth file
        outdir : str
            Output directory
        bed : str, optional
            Bed file, by default None
        pctseq : float, optional
            Percentage of sequence identity, by default 0.0
        pctsize : float, optional
            Percentage of size identity, by default 0.0
        maxsize : int, optional
            Maximum size for SV comparison, by default 50000. For CNV comparison,
            consider increasing this value or use -1 for unlimited.
        erase_outdir : bool, optional
            Erase output directory if it exists, by default True
        ignore_filter : bool, optional
            If True, ignore FILTER field (remove --passonly flag), by default False
        ignore_type : bool, optional
            If True, ignore SVTYPE when matching to truth, by default True
        skip_collapse : bool, optional
            If True, VCF collapsing step was skipped (usually for defining truthset) and
            truvari should run with -p multi
        Returns
        -------
        None
        """
        if erase_outdir and os.path.exists(outdir):
            shutil.rmtree(outdir)

        truvari_cmd = ["truvari", "bench", "-b", gt, "-c", calls, "-o", outdir]
        if ignore_type:
            truvari_cmd.append("-t")
        if not ignore_filter:
            truvari_cmd.insert(len(truvari_cmd), "--passonly")

        if bed:
            truvari_cmd.extend(["--includebed", bed])
        if skip_collapse:
            truvari_cmd.extend(["--pick", "multi"])
        truvari_cmd.extend(["--pctseq", str(pctseq)])
        truvari_cmd.extend(["--pctsize", str(pctsize)])
        truvari_cmd.extend(["--sizemax", str(maxsize)])

        self.logger.info(f"truvari command: {' '.join(truvari_cmd)}")
        self.__execute(" ".join(truvari_cmd))

    def truvari_to_dataframes(
        self,
        truvari_dir: str,
        custom_info_fields: tuple = (),
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Convert truvari report to dataframes

        Parameters
        ----------
        truvari_dir : str
            Truvari directory
        custom_info_fields : tuple, optional
            Custom info fields to read from the VCFs, in addition SVTYPE and SVLEN

        Returns
        -------
        tuple[pd.DataFrame, pd.DataFrame]
            Tuple of truvari base and calls concordance dataframes
        """

        def assign_svlen_int(df):
            return (
                df["svlen"]
                .apply(lambda x: x[0] if isinstance(x, tuple) else (x if x is not None else 0))
                .fillna(0)
                .astype(int)
            )

        info_fields_to_read: list[str] = list(("SVTYPE", "SVLEN", "MatchId") + custom_info_fields)
        df_tp_base = vcftools.get_vcf_df(pjoin(truvari_dir, "tp-base.vcf.gz"), custom_info_fields=info_fields_to_read)
        df_tp_base["svlen_int"] = assign_svlen_int(df_tp_base)
        df_tp_base["label"] = "TP"
        df_fn = vcftools.get_vcf_df(pjoin(truvari_dir, "fn.vcf.gz"), custom_info_fields=info_fields_to_read)
        df_fn["svlen_int"] = assign_svlen_int(df_fn)
        df_fn["label"] = "FN"
        df_base = pd.concat((df_tp_base, df_fn))

        df_tp_calls = vcftools.get_vcf_df(pjoin(truvari_dir, "tp-comp.vcf.gz"), custom_info_fields=info_fields_to_read)
        df_tp_calls["label"] = "TP"
        df_tp_calls["svlen_int"] = assign_svlen_int(df_tp_calls)

        df_fp = vcftools.get_vcf_df(pjoin(truvari_dir, "fp.vcf.gz"), custom_info_fields=info_fields_to_read)
        df_fp["label"] = "FP"
        df_fp["svlen_int"] = assign_svlen_int(df_fp)

        df_calls = pd.concat((df_tp_calls, df_fp))

        # Add label_type field
        # For non-TP variants, label_type equals label
        df_fn["label_type"] = df_fn["label"]
        df_fp["label_type"] = df_fp["label"]

        # For TP variants, label_type is the SVTYPE from the matching variant in the paired VCF
        # Create mapping from MatchId to SVTYPE for each TP dataframe
        if not df_tp_base.empty and "matchid" in df_tp_base.columns:
            matchid_to_svtype_calls = df_tp_calls.set_index("matchid")["svtype"].to_dict()
            df_tp_base["label_type"] = df_tp_base["matchid"].map(matchid_to_svtype_calls)
        else:
            df_tp_base["label_type"] = df_tp_base["label"]

        if not df_tp_calls.empty and "matchid" in df_tp_calls.columns:
            matchid_to_svtype_base = df_tp_base.set_index("matchid")["svtype"].to_dict()
            df_tp_calls["label_type"] = df_tp_calls["matchid"].map(matchid_to_svtype_base)
        else:
            df_tp_calls["label_type"] = df_tp_calls["label"]

        # Recreate concatenated dataframes with label_type field
        df_base = pd.concat((df_tp_base, df_fn))
        df_calls = pd.concat((df_tp_calls, df_fp))

        return df_base, df_calls

    def run_pipeline(  # noqa: PLR0913
        self,
        calls: str,
        gt: str,
        output_file_name: str,
        outdir: str,
        hcr_bed: str | None = None,
        pctseq: float = 0.0,
        pctsize: float = 0.0,
        maxsize: int = 50000,
        custom_info_fields: tuple = (),
        *,
        erase_outdir: bool = True,
        ignore_filter: bool = False,
        ignore_type: bool = True,
        skip_collapse: bool = False,
    ):
        """
        Run truvari pipeline

        Parameters
        ----------
        calls : str
            Calls file
        gt : str
            Ground truth file
        output_file_name : str
            Name of the output H5 concordance file
        outdir : str
            Output directory
        hcr_bed : str, optional
            High confidence region bed file, by default None
        pctseq : float, optional
            Percentage of sequence identity, by default 0.0
        pctsize : float, optional
            Percentage of size identity, by default 0.0
        maxsize : int, optional
            Maximum size for SV comparison, by default 50000. For CNV comparison, consider increasing this value.
        erase_outdir : bool, optional
            Erase output directory if it exists, by default True
        ignore_filter : bool, optional
            If True, ignore FILTER field (remove --passonly flag), by default False
        custom_info_fields : list[str], optional
            Custom info fields to read from the VCFs, in addition SVTYPE and SVLEN
        ignore_type : bool, optional
            If True, ignore SVTYPE when matching to truth, by default True
        skip_collapse : bool, optional
            If True, skip VCF collapsing step for calls (ground truth is always collapsed), by default False

        Returns
        -------
        None
        """

        if dirname(output_file_name) == outdir:
            raise ValueError(
                "output_file_name must not be under outdir to avoid conflicts (with running truvari bench)."
            )
        with tempfile.TemporaryDirectory(dir=dirname(output_file_name)) as workdir:
            if ignore_filter:
                self.logger.info("Ignoring FILTER field in VCFs")
                self.vu.remove_filters(input_vcf=calls, output_vcf=pjoin(workdir, "calls.nofilter.vcf.gz"))
                calls = pjoin(workdir, "calls.nofilter.vcf.gz")
                self.vu.index_vcf(calls)
            self.logger.info(f"Running truvari pipeline with calls: {calls} and gt: {gt}")
            calls_fn = calls
            tmpfiles_to_move = []

            if not skip_collapse:
                collapsed_fn = pjoin(workdir, basename(calls).replace(".vcf.gz", "_collapsed.vcf.gz"))
                self.vu.collapse_vcf(
                    calls_fn,
                    collapsed_fn,
                    bed=hcr_bed,
                    pctseq=pctseq,
                    pctsize=pctsize,
                    maxsize=maxsize,
                    ignore_filter=ignore_filter,
                    ignore_sv_type=ignore_type,
                )
                calls_fn = collapsed_fn
                tmpfiles_to_move.append(calls_fn)

                self.vu.sort_vcf(calls_fn, calls_fn.replace("_collapsed.vcf.gz", "_collapsed.sort.vcf.gz"))
                calls_fn = calls_fn.replace("_collapsed.vcf.gz", "_collapsed.sort.vcf.gz")
                tmpfiles_to_move.append(calls_fn)
                tmpfiles_to_move.append(calls_fn + ".tbi")

                self.vu.index_vcf(calls_fn)
            else:
                self.logger.info("Skipping VCF collapsing for calls")

            gt_fn = gt
            gt_collapsed_fn = pjoin(workdir, basename(gt).replace(".vcf.gz", "_collapsed.vcf.gz"))
            self.vu.collapse_vcf(
                gt_fn,
                gt_collapsed_fn,
                bed=hcr_bed,
                pctseq=pctseq,
                pctsize=pctsize,
                maxsize=maxsize,
                ignore_filter=ignore_filter,
            )
            gt_fn = gt_collapsed_fn
            tmpfiles_to_move.append(gt_fn)
            self.vu.sort_vcf(gt_fn, gt_fn.replace("_collapsed.vcf.gz", "_collapsed.sort.vcf.gz"))
            gt_fn = gt_fn.replace("_collapsed.vcf.gz", "_collapsed.sort.vcf.gz")
            tmpfiles_to_move.append(gt_fn)
            tmpfiles_to_move.append(gt_fn + ".tbi")
            self.vu.index_vcf(gt_fn)

            self.run_truvari(
                calls=calls_fn,
                gt=gt_fn,
                outdir=outdir,
                bed=hcr_bed,
                pctseq=pctseq,
                pctsize=pctsize,
                maxsize=maxsize,
                erase_outdir=erase_outdir,
                ignore_filter=ignore_filter,
                ignore_type=ignore_type,
                skip_collapse=skip_collapse,
            )
            df_base, df_calls = self.truvari_to_dataframes(outdir, custom_info_fields=custom_info_fields)
            df_base.to_hdf(output_file_name, key="base", mode="w")
            df_calls.to_hdf(output_file_name, key="calls", mode="a")
            for tmpfile in tmpfiles_to_move:
                if os.path.exists(tmpfile):
                    shutil.move(tmpfile, outdir)

        self.logger.info(f"truvari pipeline finished with calls: {calls} and gt: {gt}")


def get_parser():
    """
    Get argument parser for SVComparison

    Returns
    -------
    argparse.ArgumentParser
        Argument parser
    """
    import argparse

    parser = argparse.ArgumentParser(description="SV Comparison Pipeline")
    parser.add_argument("--calls", required=True, help="Input calls VCF file")
    parser.add_argument("--gt", required=True, help="Input ground truth VCF file")
    parser.add_argument("--output_filename", required=True, help="output h5 with concordance file")
    parser.add_argument("--outdir", required=True, help="Full path to output dir to TRUVARI results")
    parser.add_argument("--hcr_bed", help="High confidence region bed file")
    parser.add_argument("--pctseq", type=float, default=0.0, help="Percentage of sequence identity")
    parser.add_argument("--pctsize", type=float, default=0.0, help="Percentage of size identity")
    parser.add_argument(
        "--maxsize",
        type=int,
        default=50000,
        help="Maximum size for SV comparison (default: 50000). For CNV comparison, consider increasing this value.",
    )
    parser.add_argument(
        "--custom_info_fields", type=str, action="append", default=[], help="Custom info fields to read from the VCFs"
    )
    parser.add_argument(
        "--ignore_filter",
        action="store_true",
        help="Ignore FILTER field in VCF (remove --passonly flag from truvari commands)",
    )
    parser.add_argument(
        "--skip_collapse",
        action="store_true",
        help="Skip VCF collapsing step for calls (ground truth is always collapsed)",
    )
    parser.add_argument("--verbosity", default="INFO", help="Logging verbosity level")
    return parser


def run(argv):
    parser = get_parser()
    SimplePipeline.add_parse_args(parser)
    args = parser.parse_args(argv[1:])
    logger.setLevel(getattr(logging, args.verbosity))
    sp = SimplePipeline(args.fc, args.lc, debug=args.d, print_timing=True)
    pipeline = SVComparison(sp, logger)
    pipeline.run_pipeline(
        calls=args.calls,
        gt=args.gt,
        hcr_bed=args.hcr_bed,
        pctseq=args.pctseq,
        pctsize=args.pctsize,
        maxsize=args.maxsize,
        outdir=args.outdir,
        output_file_name=args.output_filename,
        custom_info_fields=tuple(args.custom_info_fields),
        ignore_filter=args.ignore_filter,
        skip_collapse=args.skip_collapse,
    )


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()



import os
import os.path
import shutil
from collections import defaultdict

import numpy as np
import pandas as pd
import pyfaidx
import pysam
import ugbio_core.concordance.flow_based_concordance as fbc
import ugbio_core.vcfbed.variant_annotation as annotation
from simppl.simple_pipeline import SimplePipeline
from ugbio_core.consts import DEFAULT_FLOW_ORDER
from ugbio_core.exec_utils import print_and_execute
from ugbio_core.logger import logger
from ugbio_core.vcf_utils import VcfUtils
from ugbio_core.vcfbed import vcftools


class VcfComparisonUtils:
    """Utilities of vcf pipeline, mostly wrappers around shell scripts

    Attributes
    ----------
    sp : SimplePipeline
        Simple pipeline object
    """

    def __init__(self, simple_pipeline: SimplePipeline | None = None, rtg_mem: str = "12G"):
        """Combines VCF in parts from GATK and indices the result

        Parameters
        ----------
        simple_pipeline : SimplePipeline, optional
            Optional SimplePipeline object for executing shell commands
        rtg_mem : str, optional
            Memory passed to RTG Tools via the RTG_MEM environment variable (default: "12G")
        """
        self.sp = simple_pipeline
        self.vu = VcfUtils()
        self.rtg_mem = rtg_mem

    def __execute(self, command: str, output_file: str | None = None):
        """Summary

        Parameters
        ----------
        command : str
            Description
        output_file : str, optional
            Description
        """
        print_and_execute(command, output_file=output_file, simple_pipeline=self.sp, module_name=__name__)

    def run_vcfeval(  # noqa PLR0913
        self,
        vcf: str,
        gt: str,
        hcr: str,
        outdir: str,
        ref_genome: str,
        ev_region: str | None = None,
        output_mode: str = "split",
        samples: str | None = None,
        *,
        erase_outdir: bool = True,
        additional_args: str = "",
        score: str = "QUAL",
        all_records: bool = False,
    ):  # pylint: disable=too-many-arguments
        """
        Run vcfeval to evaluate the concordance between two VCF files
        Parameters
        ----------
        vcf : str
            Our variant calls
        gt : str
            GIAB (or other source truth file)
        hcr : str
            High confidence regions
        outdir : str
            Output directory
        ref_genome : str
            SDF reference file
        ev_region: str, optional
            Bed file of regions to evaluate (--bed-region)
        output_mode: str, optional
            Mode of vcfeval (default - split)
        samples: str, optional
            Sample names to compare (baseline,calls)
        erase_outdir: bool, optional
            Erase the output directory if it exists before running (otherwise vcfeval crashes)
        additional_args: str, optional
            Additional arguments to pass to vcfeval
        score: str, optional
            Score field to use for producing ROC curves in VCFEVAL
        all_records: bool, optional
            Include all records in the evaluation (default - False)
        """
        if erase_outdir and os.path.exists(outdir):
            shutil.rmtree(outdir)
        cmd = [
            "rtg",
            f"RTG_MEM={self.rtg_mem}",
            "vcfeval",
            "-b",
            gt,
            "-c",
            vcf,
            "-e",
            hcr,
            "-t",
            ref_genome,
            "-m",
            output_mode,
            "--decompose",
            "-f",
            score,
            "-o",
            outdir,
        ]
        if ev_region:
            cmd += ["--bed-regions", ev_region]
        if all_records:
            cmd += ["--all-records"]
        if additional_args:
            cmd += additional_args.split()
        if samples:
            cmd += ["--sample", samples]

        logger.info(" ".join(cmd))
        return self.__execute(" ".join(cmd))

    # pylint: disable=too-many-arguments
    def run_vcfeval_concordance(  # noqa PLR0913
        self,
        input_file: str,
        truth_file: str,
        output_prefix: str,
        ref_genome: str,
        evaluation_regions: str,
        sdf_index: str | None = None,
        comparison_intervals: str | None = None,
        input_sample: str | None = None,
        truth_sample: str | None = None,
        *,
        ignore_filter: bool = False,
        mode: str = "combine",
        ignore_genotype: bool = False,
    ) -> str:
        """Run vcfeval to evaluate concordance

        Parameters
        ----------
        input_file : str
            Our variant calls
        truth_file : str
            GIAB (or other source truth file)
        output_prefix : str
            Output prefix
        ref_genome : str
            Fasta reference file
        sdf_index : str, optional
            SDF index path for the reference genome. If None, will use ref_genome + ".sdf"
        evaluation_regions: str
            Bed file of regions to evaluate (HCR)
        comparison_intervals: Optional[str]
            Picard intervals file to make the comparisons on. Default: None = all genome
        input_sample : str, optional
            Name of the sample in our input_file
        truth_sample : str, optional
            Name of the sample in the truth file
        ignore_filter : bool, optional
            Ignore status of the variant filter
        mode: str, optional
            Mode of vcfeval (default - combine)
        ignore_genotype: bool, optional
            Don't compare genotype information, only compare if allele is present in ground-truth
        Returns
        -------
        final concordance vcf file if the mode is "combine"
        otherwise - returns the output directory of vcfeval
        """

        output_dir = os.path.dirname(output_prefix)
        if sdf_index is not None:
            sdf_path = sdf_index
        else:
            sdf_path = ref_genome + ".sdf"

        if not os.path.exists(sdf_path):
            raise RuntimeError(f"Reference SDF path {sdf_path} does not exist")
        vcfeval_output_dir = os.path.join(output_dir, os.path.basename(output_prefix) + ".vcfeval_output")

        if os.path.isdir(vcfeval_output_dir):
            shutil.rmtree(vcfeval_output_dir)

        # filter the vcf to be only in the comparison_intervals.
        filtered_truth_file = os.path.join(output_dir, ".".join((os.path.basename(truth_file), "filtered", "vcf.gz")))
        if comparison_intervals is not None:
            self.vu.intersect_with_intervals(truth_file, comparison_intervals, filtered_truth_file)
        else:
            shutil.copy(truth_file, filtered_truth_file)
            self.vu.index_vcf(filtered_truth_file)

        if truth_sample is not None and input_sample is not None:
            samples = f"{truth_sample},{input_sample}"
        else:
            samples = None

        self.run_vcfeval(
            input_file,
            filtered_truth_file,
            evaluation_regions,
            vcfeval_output_dir,
            sdf_path,
            output_mode=mode,
            samples=samples,
            erase_outdir=True,
            additional_args="--squash_ploidy" if ignore_genotype else "",
            all_records=ignore_filter,
        )

        if mode == "combine":
            # fix the vcf file format
            self.fix_vcf_format(os.path.join(vcfeval_output_dir, "output"))

            # make the vcfeval output file without weird variants
            self.normalize_vcfeval_vcf(
                os.path.join(vcfeval_output_dir, "output.vcf.gz"),
                os.path.join(vcfeval_output_dir, "output.norm.vcf.gz"),
                ref_genome,
            )

            vcf_concordance_file = f"{output_prefix}.vcfeval_concordance.vcf.gz"
            # move the file to be compatible with the output file of the genotype
            # concordance
            self.__execute(f"mv {os.path.join(vcfeval_output_dir, 'output.norm.vcf.gz')} {vcf_concordance_file}")

            # generate index file for the vcf.gz file
            self.vu.index_vcf(vcf_concordance_file)
            return vcf_concordance_file
        return vcfeval_output_dir

    def normalize_vcfeval_vcf(self, input_vcf: str, output_vcf: str, ref: str) -> None:
        """Combines monoallelic rows from VCFEVAL into multiallelic
        and combines the BASE/CALL annotations together. Mostly uses `bcftools norm`,
        but since it does not aggregate the INFO tags, they are aggregated using
        `bcftools annotate`.

        Parameters
        ----------
        input_vcf: str
            Input (output.vcf.gz from VCFEVAL)
        output_vcf: str
            Input (output.vcf.gz from VCFEVAL)
        ref: str
            Reference FASTA

        Returns
        -------
        None:
            Creates output_vcf
        """

        tempdir = f"{output_vcf}_tmp"
        os.mkdir(tempdir)

        # Step1 - bcftools norm

        self.__execute(f"bcftools norm -c x -f {ref} -m+any -o {tempdir}/step1.vcf.gz -O z {input_vcf}")
        self.vu.index_vcf(f"{tempdir}/step1.vcf.gz")
        self.__execute(
            f"bcftools annotate -a {input_vcf} -c CHROM,POS,CALL,BASE -Oz \
                -o {tempdir}/step2.vcf.gz {tempdir}/step1.vcf.gz"
        )
        self.vu.index_vcf(f"{tempdir}/step2.vcf.gz")

        # Step2 - write out the annotation table. We use VariantsToTable from gatk, but remove
        # the mandatory header and replace NA by "."
        self.__execute(
            f"gatk VariantsToTable -V {input_vcf} -O {tempdir}/source.tsv -F CHROM -F POS -F REF -F CALL -F BASE"
        )
        self.__execute(
            f"gatk VariantsToTable -V {tempdir}/step2.vcf.gz -O {tempdir}/dest.tsv \
                -F CHROM -F POS -F REF -F CALL -F BASE"
        )

        # Step3 - identify lines that still need to be filled (where the CALL/BASE after the nomrmalization
        # are different from those that are expected from collapsing). This happens because plain bcftools annotate
        # requires match of reference allele that can change with normalization
        df1 = pd.read_csv(os.path.join(tempdir, "source.tsv"), sep="\t").set_index(["CHROM", "POS", "CALL"])
        df2 = pd.read_csv(os.path.join(tempdir, "dest.tsv"), sep="\t").set_index(["CHROM", "POS", "CALL"])

        difflines = df1.loc[df1.index.difference(df2.index)].reset_index().fillna(".")[["CHROM", "POS", "CALL"]]
        difflines.to_csv(os.path.join(tempdir, "step3.call.tsv"), sep="\t", header=False, index=False)

        df1 = pd.read_csv(os.path.join(tempdir, "source.tsv"), sep="\t").set_index(["CHROM", "POS", "BASE"])
        df2 = pd.read_csv(os.path.join(tempdir, "dest.tsv"), sep="\t").set_index(["CHROM", "POS", "BASE"])

        difflines = df1.loc[df1.index.difference(df2.index)].reset_index().fillna(".")[["CHROM", "POS", "BASE"]]
        difflines.to_csv(os.path.join(tempdir, "step3.base.tsv"), sep="\t", header=False, index=False)

        # Step 4 - annoate with the additional tsvs
        self.__execute(f"bgzip {tempdir}/step3.call.tsv")
        self.__execute(f"bgzip {tempdir}/step3.base.tsv")
        self.__execute(f"tabix -s1 -e2 -b2 {tempdir}/step3.call.tsv.gz")
        self.__execute(f"tabix -s1 -e2 -b2 {tempdir}/step3.base.tsv.gz")
        self.__execute(
            f"bcftools annotate -c CHROM,POS,CALL -a {tempdir}/step3.call.tsv.gz \
                -Oz -o {tempdir}/step4.vcf.gz {tempdir}/step2.vcf.gz"
        )
        self.vu.index_vcf(f"{tempdir}/step4.vcf.gz")
        self.__execute(
            f"bcftools annotate -c CHROM,POS,CALL -a {tempdir}/step3.base.tsv.gz \
                -Oz -o {output_vcf} {tempdir}/step4.vcf.gz"
        )
        self.vu.index_vcf(output_vcf)
        # shutil.rmtree(tempdir)

    def fix_vcf_format(self, output_prefix: str):
        """Legacy function to fix the PS field format in the old GIAB truth sets. The function overwrites the input file

        Parameters
        ----------
        output_prefix : str
            Prefix of the input and the output file (without the .vcf.gz)
        """
        self.__execute(f"gunzip -f {output_prefix}.vcf.gz")
        with open(f"{output_prefix}.vcf", encoding="utf-8") as input_file_handle:
            with open(f"{output_prefix}.tmp", "w", encoding="utf-8") as output_file_handle:
                for line in input_file_handle:
                    if line.startswith("##FORMAT=<ID=PS"):
                        output_file_handle.write(line.replace("Type=Integer", "Type=String"))
                    else:
                        output_file_handle.write(line)
        self.__execute(f"mv {output_file_handle.name} {input_file_handle.name}")
        self.__execute(f"bgzip {input_file_handle.name}")
        self.vu.index_vcf(f"{input_file_handle.name}.gz")

    def transform_hom_calls_to_het_calls(self, input_file_calls: str, output_file_calls: str) -> None:
        """Reverse homozygous reference calls in deepVariant to filtered heterozygous so that max recall can be
        calculated

        Parameters
        ----------
        input_file_calls : str
            Input file name
        output_file_calls : str
            Output file name
        """

        with pysam.VariantFile(input_file_calls) as input_file:
            with pysam.VariantFile(output_file_calls, "w", header=input_file.header) as output_file:
                for rec in input_file:
                    if (
                        rec.samples[0]["GT"] == (0, 0)
                        or rec.samples[0]["GT"] == (None, None)
                        and "PASS" not in rec.filter
                    ):
                        rec.samples[0]["GT"] = (0, 1)
                    output_file.write(rec)
        self.vu.index_vcf(output_file_calls)


def _fix_errors(df: pd.DataFrame) -> pd.DataFrame:
    """Parses dataframe generated from the VCFEVAL concordance VCF and prepares it for
     classify/classify_gt functions that only consider the genotypes of the call and the base
    only rather than at the classification that VCFEVAL produced

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe

    Returns
    -------
    pd.DataFrame
        Output dataframe,
    """

    # remove genotypes of variants that were filtered out and thus are false negatives
    # (VCFEVAL outputs UG genotype for ignored genotypes too and they are classified downstream
    # as true positives if we do not make this modification)
    fix_tp_fn_loc = (df["call"] == "IGN") & ((df["base"] == "FN") | (df["base"] == "FN_CA"))
    replace = df.loc[fix_tp_fn_loc, "gt_ultima"].apply(lambda x: (None,))
    df.loc[replace.index, "gt_ultima"] = replace

    # fix all the places in which vcfeval returns a good result, but the genotype is not adequate
    # in these cases we change the genotype of the gt to be adequate with the classify function as follow:
    # (TP,TP), (TP,None) - should put the values of ultima in the gt
    df.loc[(df["call"] == "TP") & ((df["base"] == "TP") | (df["base"].isna())), "gt_ground_truth"] = df[
        (df["call"] == "TP") & ((df["base"] == "TP") | (df["base"].isna()))
    ]["gt_ultima"]

    # (FP_CA,FN_CA), (FP_CA,None) - Fake a genotype from ultima such that one of the alleles is the same (and only one)
    df.loc[(df["call"] == "FP_CA") & ((df["base"] == "FN_CA") | (df["base"].isna())), "gt_ground_truth"] = df[
        (df["call"] == "FP_CA") & ((df["base"] == "FN_CA") | (df["base"].isna()))
    ]["gt_ultima"].apply(
        lambda x: ((x[0], x[0]) if (len(x) < 2 or (x[1] == 0)) else ((x[1], x[1]) if (x[0] == 0) else (x[0], 0)))  # noqa PLR2004
    )
    return df


def __map_variant_to_dict(variant: pysam.VariantRecord) -> defaultdict:
    """Converts a line from vcfeval concordance VCF to a dictionary. The following fields are extracted
    call genotype, base genotype, qual, chromosome, position, ref, alt and all values from the INFO column

    Parameters
    ----------
    variant : pysam.VariantRecord
        VCF record

    Returns
    -------
    defaultdict
        Output dictionary
    """
    call_sample_ind = 1
    gtr_sample_ind = 0

    return defaultdict(
        lambda: None,
        variant.info.items()
        + [
            ("GT_ULTIMA", variant.samples[call_sample_ind]["GT"]),
            ("GT_GROUND_TRUTH", variant.samples[gtr_sample_ind]["GT"]),
            ("QUAL", variant.qual),
            ("CHROM", variant.chrom),
            ("POS", variant.pos),
            ("REF", variant.ref),
            ("ALLELES", variant.alleles),
        ],
    )


def vcf2concordance(  # noqa PLR0915 C901
    raw_calls_file: str,
    concordance_file: str,
    chromosome: str | None = None,
    scoring_field: str | None = None,
) -> pd.DataFrame:
    """Generates concordance dataframe

    Parameters
    ----------
    raw_calls_file : str
        File with GATK calls (.vcf.gz)
    concordance_file : str
        GenotypeConcordance/VCFEVAL output file (.vcf.gz)
    chromosome: str
        Fetch a specific chromosome (Default - all)
    scoring_field : str, optional
        The name of the INFO field that is used to score the variants.
        This value replaces the TREE_SCORE in the output data frame.
        When None TREE_SCORE is not replaced (default: None)

    Returns
    -------
    pd.DataFrame
        Concordance dataframe
    """

    if chromosome is None:
        concord_vcf = pysam.VariantFile(concordance_file)
    else:
        concord_vcf = pysam.VariantFile(concordance_file).fetch(chromosome)

    def call_filter(x):
        # Remove variants that were ignored (either outside of comparison intervals or
        # filtered out).
        return not (
            (x["CALL"] in {"IGN", "OUT"} and x["BASE"] is None)
            or (x["CALL"] in {"IGN", "OUT"} and x["BASE"] in {"IGN", "OUT"})
            or (x["CALL"] is None and x["BASE"] in {"IGN", "OUT"})
        )

    concord_vcf_extend = filter(call_filter, (__map_variant_to_dict(variant) for variant in concord_vcf))

    columns = [
        "CHROM",
        "POS",
        "QUAL",
        "REF",
        "ALLELES",
        "GT_ULTIMA",
        "GT_GROUND_TRUTH",
        "SYNC",
        "CALL",
        "BASE",
        "STR",
        "RU",
        "RPA",
    ]
    column_names = [x.lower() for x in columns]
    concordance_df = pd.DataFrame([[x[y] for y in columns] for x in concord_vcf_extend], columns=column_names)

    # make the gt_ground_truth compatible with GC
    concordance_df["gt_ground_truth"] = concordance_df["gt_ground_truth"].map(
        lambda x: (None, None) if x == (None,) else x
    )

    concordance_df["indel"] = concordance_df["alleles"].apply(lambda x: len({len(y) for y in x}) > 1)
    concordance_df = _fix_errors(concordance_df)

    def classify(x: defaultdict | pd.Series | dict) -> str:
        """Classify a record as true positive / false positive / false negative by matching the alleles.
        TP will be called if some of the alleles match

        Parameters
        ----------
        x : defaultdict, pd.Series or dict
            Input record

        Returns
        -------
        str
            classification
        """
        if x["gt_ultima"] == (None, None) or x["gt_ultima"] == (None,):
            return "fn"

        if x["gt_ground_truth"] == (None, None) or x["gt_ground_truth"] == (None,):
            return "fp"

        # If both gt_ultima and gt_ground_truth are not none:
        set_gtr = set(x["gt_ground_truth"]) - {0}
        set_ultima = set(x["gt_ultima"]) - {0}

        if len(set_gtr & set_ultima) > 0:
            return "tp"

        if len(set_ultima - set_gtr) > 0:
            return "fp"

        # If it is not tp or fp, then return fn:
        return "fn"

    concordance_df["classify"] = concordance_df.apply(classify, axis=1, result_type="reduce")

    def classify_gt(x: defaultdict | pd.Series | dict) -> str:
        """Classify a record as true positive / false negative / false positive. True positive requires
        match in the alleles and genotypes

        Parameters
        ----------
        x : defaultdict, pd.Series or dict
            Input record

        Returns
        -------
        str
            Classification
        """
        n_ref_gtr = len([y for y in x["gt_ground_truth"] if y == 0])
        n_ref_ultima = len([y for y in x["gt_ultima"] if y == 0])

        if x["gt_ultima"] == (None, None) or x["gt_ultima"] == (None,):
            return "fn"
        if x["gt_ground_truth"] == (None, None) or x["gt_ground_truth"] == (None,):
            return "fp"
        if n_ref_gtr < n_ref_ultima:
            return "fn"
        if n_ref_gtr > n_ref_ultima:
            return "fp"
        if x["gt_ultima"] != x["gt_ground_truth"]:
            return "fp"
        # If not fn or fp due to the reasons above:
        return "tp"

    concordance_df["classify_gt"] = concordance_df.apply(classify_gt, axis=1, result_type="reduce")
    concordance_df.loc[
        (concordance_df["classify_gt"] == "tp") & (concordance_df["classify"] == "fp"),
        "classify_gt",
    ] = "fp"

    # cases where we called wrong allele and then filtered out - are false negatives, not false positives
    called_fn = (concordance_df["base"] == "FN") | (concordance_df["base"] == "FN_CA")
    marked_fp = concordance_df["classify"] == "fp"
    concordance_df.loc[called_fn & marked_fp, "classify"] = "fn"
    marked_fp = concordance_df["classify_gt"] == "fp"
    concordance_df.loc[called_fn & marked_fp, "classify_gt"] = "fn"
    concordance_df.index = pd.Index(list(zip(concordance_df.chrom, concordance_df.pos, strict=False)))
    original = vcftools.get_vcf_df(raw_calls_file, chromosome=chromosome, scoring_field=scoring_field)

    concordance_df = concordance_df.drop("qual", axis=1)

    drop_candidates = ["chrom", "pos", "alleles", "indel", "ref", "str", "ru", "rpa"]
    if original.shape[0] > 0:
        concordance = concordance_df.join(
            original.drop(
                [x for x in drop_candidates if x in original.columns and x in concordance_df.columns],
                axis=1,
            )
        )
    else:
        concordance = concordance_df.copy()

        tmp = original.drop(
            [x for x in drop_candidates if x in original.columns and x in concordance_df.columns],
            axis=1,
        )
        for t in tmp.columns:
            concordance[t] = None

    only_ref = concordance["alleles"].apply(len) == 1
    concordance = concordance[~only_ref]

    # Marking as false negative the variants that appear in concordance but not in the
    # original VCF (even if they do show some genotype)
    missing_variants = concordance.index.difference(original.index)
    logger.info("Identified %i variants missing in the input VCF", len(missing_variants))
    missing_variants_non_fn = concordance.loc[missing_variants].query("classify!='fn'").index
    logger.warning(
        "Identified %i variants missing in the input VCF and not marked false negatives",
        len(missing_variants_non_fn),
    )
    concordance.loc[missing_variants_non_fn, "classify"] = "fn"
    concordance.loc[missing_variants_non_fn, "classify_gt"] = "fn"

    return concordance


def annotate_concordance(
    concordance_df: pd.DataFrame,
    fasta: str,
    bw_high_quality: list[str] | None = None,
    bw_all_quality: list[str] | None = None,
    annotate_intervals: list[str] | None = None,
    flow_order: str | None = DEFAULT_FLOW_ORDER,
) -> tuple[pd.DataFrame, list]:
    """Annotates concordance data with information about SNP/INDELs and motifs

    Parameters
    ----------
    concordance_df : pd.DataFrame
        Concordance dataframe
    fasta : str
        Indexed FASTA of the reference genome
    bw_high_quality : list[str], optional
        Coverage bigWig file from high mapq reads  (Optional)
    bw_all_quality : list[str], optional
        Coverage bigWig file from all mapq reads  (Optional)
    annotate_intervals : list[str], optional
        Interval files for annotation
    flow_order : str, optional
        Flow order

    Returns
    -------
    pd.DataFrame
        Annotated dataframe
    list
        List of the names of the annotations
    """

    if annotate_intervals is None:
        annotate_intervals = []

    logger.info("Marking SNP/INDEL")
    concordance_df = annotation.classify_indel(concordance_df)
    logger.info("Marking H-INDEL")
    concordance_df = annotation.is_hmer_indel(concordance_df, fasta)
    logger.info("Marking motifs")
    concordance_df = annotation.get_motif_around(concordance_df, 5, fasta)
    logger.info("Marking GC content")
    concordance_df = annotation.get_gc_content(concordance_df, 10, fasta)
    if bw_all_quality is not None and bw_high_quality is not None:
        logger.info("Calculating coverage")
        concordance_df = annotation.get_coverage(concordance_df, bw_high_quality, bw_all_quality)
    annots = []
    for annotation_file in annotate_intervals:
        logger.info("Annotating intervals")
        concordance_df, annot = annotation.annotate_intervals(concordance_df, annotation_file)
        annots.append(annot)
    logger.debug("Filling filter column")  # debug since not interesting step
    concordance_df = annotation.fill_filter_column(concordance_df)
    if flow_order is not None:
        concordance_df = annotation.annotate_cycle_skip(concordance_df, flow_order=flow_order)
    return concordance_df, annots


def reinterpret_variants(
    concordance_df: pd.DataFrame,
    reference_fasta: str,
    *,
    ignore_low_quality_fps: bool = False,
) -> pd.DataFrame:
    """Reinterprets the variants by comparing the variant to the ground truth in flow space

    Parameters
    ----------
    concordance_df : pd.DataFrame
        Input dataframe
    reference_fasta : str
        Indexed FASTA
    ignore_low_quality_fps : bool, optional
        Should the low quality false positives be ignored in reinterpretation (True for mutect, default False)

    Returns
    -------
    pd.DataFrame
        Reinterpreted dataframe

    See Also
    --------
    `flow_based_concordance.py`
    """
    logger.info("Variants reinterpret")
    concordance_df_result = pd.DataFrame()
    fasta = pyfaidx.Fasta(reference_fasta, build_index=False, rebuild=False)
    for contig in concordance_df["chrom"].unique():
        concordance_df_contig = concordance_df.loc[concordance_df["chrom"] == contig]
        input_dict = _get_locations_to_work_on(concordance_df_contig, ignore_low_quality_fps=ignore_low_quality_fps)
        concordance_df_contig = fbc.reinterpret_variants(concordance_df_contig, input_dict, fasta)
        concordance_df_result = pd.concat([concordance_df_result, concordance_df_contig])
    return concordance_df_result


def _get_locations_to_work_on(input_df: pd.DataFrame, *, ignore_low_quality_fps: bool = False) -> dict:
    """Dictionary of locations in the dataframe that we care about

    Parameters
    ----------
    input_df : pd.DataFrame
        Input
    ignore_low_quality_fps : bool, optional
        Should we ignore the low quality false positives

    Returns
    -------
    dict
        locations dictionary split between fps/fns/tps etc.

    """
    filtered_df = vcftools.FilterWrapper(input_df)
    fps = filtered_df.reset().get_fp().get_df()
    if "tree_score" in fps.columns and fps["tree_score"].dtype == np.float64 and ignore_low_quality_fps:
        cutoff = fps.tree_score.quantile(0.80)
        fps = fps.query(f"tree_score > {cutoff}")
    fns = filtered_df.reset().get_df().query('classify=="fn"')
    tps = filtered_df.reset().get_tp().get_df()
    gtr = (
        filtered_df.reset()
        .get_df()
        .loc[filtered_df.get_df()["gt_ground_truth"].apply(lambda x: x not in [(None, None), (None,)])]
        .copy()
    )
    gtr = gtr.sort_values("pos")
    ugi = (
        filtered_df.reset()
        .get_df()
        .loc[filtered_df.get_df()["gt_ultima"].apply(lambda x: x not in [(None, None), (None,)])]
        .copy()
    )
    ugi = ugi.sort_values("pos")

    pos_fps = np.array(fps.pos)
    pos_gtr = np.array(gtr.pos)
    pos_ugi = np.array(ugi.pos)
    pos_fns = np.array(fns.pos)

    result = {
        "fps": fps,
        "fns": fns,
        "tps": tps,
        "gtr": gtr,
        "ugi": ugi,
        "pos_fps": pos_fps,
        "pos_gtr": pos_gtr,
        "pos_ugi": pos_ugi,
        "pos_fns": pos_fns,
    }

    return result



from __future__ import annotations

import shutil
from os.path import basename, dirname, splitext
from os.path import join as pjoin

from ugbio_core.vcf_utils import VcfUtils
from ugbio_core.vcfbed.interval_file import IntervalFile

from ugbio_comparison.vcf_comparison_utils import VcfComparisonUtils


class ComparisonPipeline:  # pylint: disable=too-many-instance-attributes
    """
    Run comparison between the two sets of calls: input_prefix and truth_file. Creates
    a combined call file and a concordance VCF by either of the concordance tools
    """

    # pylint: disable=too-many-arguments
    def __init__(  # noqa: PLR0913
        self,
        vcu: VcfComparisonUtils,
        vu: VcfUtils,
        input_prefix: str,
        truth_file: str,
        cmp_intervals: IntervalFile,
        highconf_intervals: IntervalFile,
        ref_genome: str,
        call_sample: str,
        truth_sample: str,
        output_file_name: str,
        sdf_index: str | None = None,
        output_suffix: str | None = None,
        *,
        ignore_filter: bool = False,
        revert_hom_ref: bool = False,
    ):
        """
        Parameters
        ----------
        vcu: VcfComparisonUtils
            VcfComparisonUtils object for executing common functions via unix shell
        vu: VcfUtils
            VcfUtils object for VCF utility functions
        input_prefix : str
            Input prefix for the vcf. The script will look for <input_prefix>.vcf.gz
        truth_file : str, optional
            Truth calls file
        cmp_intervals : vcf_pipeline_utils.IntervalFile, optional
            interval_list file over which to do comparison (e.g. chr9)
        highconf_intervals : IntervalFile
            high confidence intervals for the ground truth
        ref_genome : str
            Reference genome FASTA
        sdf_index : str, optional
            VCFEVAL SDF index for the reference genome. If None, will use ref_genome + ".sdf"
        call_sample : str, optional
            Name of the calls sample
        truth_sample : str, optional
            Name of the truth sample
        output_file_name: str
            Name of the output file - will determine the name of the vcfeval output directory.
            The output_file_name should include the output_dir and output_dir and output_file_name
            are mutually exclusive
        output_suffix : str, optional
            Suffix for the output file name (e.g. chr9) -
            otherwise the output file nams are starting with the input prefix
        ignore_filter : bool, optional
            Should the filter status **of calls only** be ignored. Filter status of truth is always
            taken into account
        revert_hom_ref : bool, optional
            Should the hom ref calls filtered out be reverted (deepVariant only)
        """
        self.vcu = vcu
        self.vu = vu
        self.input_prefix = input_prefix
        self.truth_file = truth_file
        self.cmp_intervals = cmp_intervals
        self.ref_genome = ref_genome
        self.sdf_index = sdf_index
        self.highconf_intervals = highconf_intervals
        self.call_sample = call_sample
        self.truth_sample = truth_sample
        self.output_file_name = output_file_name
        self.ignore_filter = ignore_filter
        self.revert_hom_ref = revert_hom_ref
        self.output_suffix = output_suffix

        self.output_dir = dirname(output_file_name)
        self.output_prefix = splitext(output_file_name)[0]
        self.input_prefix_basename = basename(input_prefix)
        self.output_suffix = "" if not self.output_suffix else "." + self.output_suffix

    def run(self) -> tuple[str, str]:
        """
        Returns
        -------
        high_conf_calls_vcf, high_conf_concordance_vcf: Tuple[str, str]
        """

        combined_fn = self.input_prefix + ".vcf.gz"
        revert_fn = self.__revert_hom_ref(combined_fn)
        select_intervals_fn = self.__select_comparison_intervals(revert_fn)

        concordance_vcf = self.vcu.run_vcfeval_concordance(
            input_file=select_intervals_fn,
            truth_file=self.truth_file,
            output_prefix=self.output_prefix,
            ref_genome=self.ref_genome,
            sdf_index=self.sdf_index,
            evaluation_regions=str(self.highconf_intervals.as_bed_file()),
            comparison_intervals=self.cmp_intervals.as_bed_file(),
            input_sample=self.call_sample,
            truth_sample=self.truth_sample,
            ignore_filter=self.ignore_filter,
            mode="combine",
        )
        annotated_concordance_vcf = self.vu.annotate_tandem_repeats(concordance_vcf, self.ref_genome)

        high_conf_calls_vcf = select_intervals_fn.replace("vcf.gz", "highconf.vcf.gz")
        self.vu.intersect_with_intervals(
            select_intervals_fn, str(self.highconf_intervals.as_interval_list_file()), high_conf_calls_vcf
        )

        high_conf_concordance_vcf = annotated_concordance_vcf.replace("vcf.gz", "highconf.vcf.gz")

        self.vu.intersect_with_intervals(
            annotated_concordance_vcf, str(self.highconf_intervals.as_interval_list_file()), high_conf_concordance_vcf
        )
        return high_conf_calls_vcf, high_conf_concordance_vcf

    def __revert_hom_ref(self, combined_fn):
        revert_fn = pjoin(self.output_dir, self.input_prefix_basename + f"{self.output_suffix}.rev.hom.ref.vcf.gz")
        if self.revert_hom_ref:
            self.vcu.transform_hom_calls_to_het_calls(combined_fn, revert_fn)
        else:
            shutil.copy(combined_fn, revert_fn)
            shutil.copy(".".join((combined_fn, "tbi")), ".".join((revert_fn, "tbi")))
        return revert_fn

    def __select_comparison_intervals(self, revert_fn):
        select_intervals_fn = pjoin(self.output_dir, self.input_prefix_basename + f"{self.output_suffix}.intsct.vcf.gz")
        if not self.cmp_intervals.is_none():
            self.vu.intersect_with_intervals(
                revert_fn, str(self.cmp_intervals.as_interval_list_file()), select_intervals_fn
            )
        else:
            shutil.copy(revert_fn, select_intervals_fn)
            self.vu.index_vcf(select_intervals_fn)
        return select_intervals_fn



#!/env/python
# Copyright 2022 Ultima Genomics Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# DESCRIPTION
#    Compare UG callset to ground truth (preferentially using VCFEVAL as an engine)
# CHANGELOG in reverse chronological order

from __future__ import annotations

import argparse
import logging
import os
import sys
from os.path import dirname
from shutil import copyfile

import pandas as pd
import pysam
from joblib import Parallel, delayed
from simppl.simple_pipeline import SimplePipeline
from tqdm import tqdm
from ugbio_core.bed_utils import BedUtils
from ugbio_core.consts import DEFAULT_FLOW_ORDER
from ugbio_core.h5_utils import read_hdf
from ugbio_core.logger import logger
from ugbio_core.vcf_utils import VcfUtils
from ugbio_core.vcfbed import vcftools
from ugbio_core.vcfbed.interval_file import IntervalFile

from ugbio_comparison.comparison_pipeline import ComparisonPipeline
from ugbio_comparison.vcf_comparison_utils import (
    VcfComparisonUtils,
    annotate_concordance,
    reinterpret_variants,
    vcf2concordance,
)

MIN_CONTIG_LENGTH = 100000


def _contig_concordance_annotate_reinterpretation(  # noqa: PLR0913
    # pylint: disable=too-many-arguments
    raw_calls_vcf,
    concordance_vcf,
    contig,
    reference,
    bw_high_quality,
    bw_all_quality,
    annotate_intervals,
    flow_order,
    base_name_outputfile,
    enable_reinterpretation,
    ignore_low_quality_fps,
    scoring_field,
):
    logger.info("Reading %s", contig)
    concordance = vcf2concordance(
        raw_calls_vcf,
        concordance_vcf,
        contig,
        scoring_field=scoring_field,
    )

    annotated_concordance, _ = annotate_concordance(
        concordance,
        reference,
        bw_high_quality,
        bw_all_quality,
        annotate_intervals,
        flow_order=flow_order,
    )

    if enable_reinterpretation:
        annotated_concordance = reinterpret_variants(
            annotated_concordance, reference, ignore_low_quality_fps=ignore_low_quality_fps
        )
    logger.debug("%s: %s", contig, annotated_concordance.shape)
    annotated_concordance.to_hdf(f"{base_name_outputfile}{contig}.h5", key=contig)


def get_parser() -> argparse.ArgumentParser:
    ap_var = argparse.ArgumentParser(prog="run_comparison_pipeline.py", description="Compare VCF to ground truth")
    ap_var.add_argument("--input_prefix", help="Prefix of the input file", required=True, type=str)
    ap_var.add_argument("--output_file", help="Output h5 file", required=True, type=str)
    ap_var.add_argument(
        "--output_interval",
        help="Output bed file of intersected intervals",
        required=True,
        type=str,
    )
    ap_var.add_argument("--gtr_vcf", help="Ground truth VCF file", required=True, type=str)
    ap_var.add_argument(
        "--cmp_intervals",
        help="Ranges on which to perform comparison (bed/interval_list)",
        required=False,
        type=str,
        default=None,
    )
    ap_var.add_argument(
        "--highconf_intervals",
        help="High confidence intervals (bed/interval_list)",
        required=True,
        type=str,
    )
    ap_var.add_argument(
        "--annotate_intervals",
        help="interval files for annotation (multiple possible)",
        required=False,
        type=str,
        default=None,
        action="append",
    )
    ap_var.add_argument("--reference", help="Reference genome", required=True, type=str)
    ap_var.add_argument("--reference_dict", help="Reference genome dictionary", required=False, type=str)
    ap_var.add_argument(
        "--reference_sdf",
        help="VCFEVAL SDF index for the reference genome, in case difference from reference.sdf",
        required=False,
        type=str,
        default=None,
    )
    ap_var.add_argument(
        "--coverage_bw_high_quality",
        help="BigWig file with coverage only on high mapq reads",
        required=False,
        default=None,
        type=str,
        action="append",
    )
    ap_var.add_argument(
        "--coverage_bw_all_quality",
        help="BigWig file with coverage on all mapq reads",
        required=False,
        default=None,
        type=str,
        action="append",
    )
    ap_var.add_argument(
        "--call_sample_name",
        help="Name of the call sample",
        required=True,
        default="sm1",
    )
    ap_var.add_argument("--truth_sample_name", help="Name of the truth sample", required=True)
    ap_var.add_argument("--ignore_filter_status", help="Ignore variant filter status", action="store_true")
    ap_var.add_argument(
        "--revert_hom_ref",
        help="For DeepVariant callsets - revert filtered hom_ref to het_ref for max recall calculation",
        action="store_true",
    )
    ap_var.add_argument(
        "--scoring_field",
        help="The pipeline expects a TREE_SCORE column in order to score the variants. If another field is \
        provided via scoring_field then its values will be copied to the TREE_SCORE column",
        required=False,
        default=None,
        type=str,
    )
    ap_var.add_argument(
        "--flow_order",
        type=str,
        help="Sequencing flow order (4 cycle)",
        required=False,
        default=DEFAULT_FLOW_ORDER,
    )
    ap_var.add_argument(
        "--output_suffix",
        help="Add suffix to the output file",
        required=False,
        default="",
        type=str,
    )
    ap_var.add_argument(
        "--enable_reinterpretation",
        help="Enable variant reinterpretation in flow space (disabled by default)",
        action="store_true",
    )
    ap_var.add_argument(
        "--special_chromosome",
        help="The chromosome that would be used for the \
        'concordance' dataframe (whole genome mode only)",
        default="chr9",
    )
    ap_var.add_argument("--is_mutect", help="Are the VCFs output of Mutect (false)", action="store_true")
    ap_var.add_argument("--n_jobs", help="n_jobs of parallel on contigs", type=int, default=-1)
    ap_var.add_argument(
        "--use_tmpdir", help="Should temporary files be stored in temporary directory", action="store_true"
    )
    ap_var.add_argument(
        "--verbosity",
        help="Verbosity: ERROR, WARNING, INFO, DEBUG",
        required=False,
        default="INFO",
    )
    ap_var.add_argument(
        "--rtg_mem",
        help='Memory for RTG Tools via RTG_MEM (e.g. "12G", "24G")',
        required=False,
        default="12G",
        type=str,
    )
    return ap_var


def _setup_pipeline_and_utilities(args) -> tuple[SimplePipeline, VcfComparisonUtils, VcfUtils]:
    """Setup pipeline and utility objects."""
    logger.setLevel(getattr(logging, args.verbosity))
    sp = SimplePipeline(args.fc, args.lc, debug=args.d, print_timing=True)
    vcu = VcfComparisonUtils(sp, rtg_mem=args.rtg_mem)
    vu = VcfUtils(sp)
    return sp, vcu, vu


def _setup_interval_files(sp, args) -> tuple[IntervalFile, IntervalFile]:
    """Setup interval file objects."""
    # use tmpdir consistent with comparison_pipeline
    if args.use_tmpdir:
        scratchdir = dirname(args.output_file)
    else:
        scratchdir = False  # do not use tmpdir

    cmp_intervals = IntervalFile(sp, args.cmp_intervals, args.reference, args.reference_dict, scratchdir=scratchdir)
    highconf_intervals = IntervalFile(
        sp, args.highconf_intervals, args.reference, args.reference_dict, scratchdir=scratchdir
    )

    return cmp_intervals, highconf_intervals


def _intersect_intervals_and_save_args(vu, cmp_intervals, highconf_intervals, args) -> None:
    """Intersect intervals and save arguments to output file."""
    # intersect intervals and output as a bed file
    if cmp_intervals.is_none():  # interval of highconf_intervals
        logger.info(f"copy {args.highconf_intervals} to {args.output_interval}")
        copyfile(str(highconf_intervals.as_bed_file()), args.output_interval)
    else:
        BedUtils().intersect_bed_files(
            str(cmp_intervals.as_bed_file()), str(highconf_intervals.as_bed_file()), args.output_interval
        )

    # Save input arguments
    args_dict = {k: str(vars(args)[k]) for k in vars(args)}
    pd.DataFrame(args_dict, index=[0]).to_hdf(args.output_file, key="input_args")


def _create_comparison_pipeline(vcu, vu, cmp_intervals, highconf_intervals, args) -> ComparisonPipeline:
    """Create and configure the comparison pipeline."""
    return ComparisonPipeline(
        vcu=vcu,
        vu=vu,
        input_prefix=args.input_prefix,
        truth_file=args.gtr_vcf,
        cmp_intervals=cmp_intervals,
        highconf_intervals=highconf_intervals,
        ref_genome=args.reference,
        sdf_index=args.reference_sdf,
        call_sample=args.call_sample_name,
        truth_sample=args.truth_sample_name,
        output_file_name=args.output_file,
        output_suffix=args.output_suffix,
        ignore_filter=args.ignore_filter_status,
        revert_hom_ref=args.revert_hom_ref,
    )


def _process_single_interval_concordance(raw_calls_vcf, concordance_vcf, args) -> None:
    """Process concordance for single interval file."""
    concordance_df = vcf2concordance(
        raw_calls_vcf,
        concordance_vcf,
        scoring_field=args.scoring_field,
    )
    annotated_concordance_df, _ = annotate_concordance(
        concordance_df,
        args.reference,
        args.coverage_bw_high_quality,
        args.coverage_bw_all_quality,
        args.annotate_intervals,
        flow_order=args.flow_order,
    )

    if args.enable_reinterpretation:
        annotated_concordance_df = reinterpret_variants(
            annotated_concordance_df,
            args.reference,
            ignore_low_quality_fps=args.is_mutect,
        )

    annotated_concordance_df.to_hdf(args.output_file, key="concordance", mode="a")
    # hack until we totally remove chr9
    annotated_concordance_df.to_hdf(args.output_file, key="comparison_result", mode="a")
    vcftools.bed_files_output(
        annotated_concordance_df,
        args.output_file,
        mode="w",
        create_gt_diff=(not args.is_mutect),
    )


def _get_filtered_contigs(raw_calls_vcf) -> list[str]:
    """Get list of contigs filtered by minimum length."""
    with pysam.VariantFile(raw_calls_vcf) as variant_file:
        # we filter out short contigs to prevent huge files
        filtered_contigs = []
        for contig_name in variant_file.header.contigs:
            contig = variant_file.header.contigs[contig_name]
            if contig.length is not None and contig.length > MIN_CONTIG_LENGTH:
                filtered_contigs.append(str(contig_name))
        return filtered_contigs


def _process_whole_genome_concordance(raw_calls_vcf, concordance_vcf, args) -> None:
    """Process concordance for whole genome (per chromosome)."""
    contigs = _get_filtered_contigs(raw_calls_vcf)
    base_name_outputfile = os.path.splitext(args.output_file)[0]

    # Process each contig in parallel
    Parallel(n_jobs=args.n_jobs, max_nbytes=None)(
        delayed(_contig_concordance_annotate_reinterpretation)(
            raw_calls_vcf,
            concordance_vcf,
            contig,
            args.reference,
            args.coverage_bw_high_quality,
            args.coverage_bw_all_quality,
            args.annotate_intervals,
            args.flow_order,
            base_name_outputfile,
            args.enable_reinterpretation,
            args.is_mutect,
            args.scoring_field,
        )
        for contig in tqdm(contigs)
    )

    _merge_contig_results(contigs, base_name_outputfile, args)


def _merge_contig_results(contigs, base_name_outputfile, args) -> None:
    """Merge temporary H5 files and generate bed files."""
    # find columns and set the same header for empty dataframes
    df_columns = None
    for contig in contigs:
        h5_temp = read_hdf(f"{base_name_outputfile}{contig}.h5", key=str(contig))
        if h5_temp.shape == (0, 0):  # empty dataframes are dropped to save space
            continue
        df_columns = pd.DataFrame(columns=h5_temp.columns)
        break

    # Merge temp files into main output file
    for contig in contigs:
        h5_temp = read_hdf(f"{base_name_outputfile}{contig}.h5", key=str(contig))
        if h5_temp.shape == (0, 0):  # empty dataframes get default columns
            h5_temp = pd.concat((h5_temp, df_columns), axis=1)
        h5_temp.to_hdf(args.output_file, mode="a", key=str(contig))
        if contig == args.special_chromosome:
            h5_temp.to_hdf(args.output_file, mode="a", key="concordance")
        os.remove(f"{base_name_outputfile}{contig}.h5")

    # Generate bed files
    write_mode = "w"
    for contig in contigs:
        annotated_concordance_df = read_hdf(args.output_file, key=str(contig))
        vcftools.bed_files_output(
            annotated_concordance_df,
            args.output_file,
            mode=write_mode,
            create_gt_diff=(not args.is_mutect),
        )
        write_mode = "a"


def run(argv: list[str]):
    """Concordance between VCF and ground truth"""
    parser = get_parser()
    SimplePipeline.add_parse_args(parser)
    args = parser.parse_args(argv[1:])

    # Setup pipeline and utilities
    sp, vcu, vu = _setup_pipeline_and_utilities(args)

    # Setup interval files
    cmp_intervals, highconf_intervals = _setup_interval_files(sp, args)

    # Intersect intervals and save arguments
    _intersect_intervals_and_save_args(vu, cmp_intervals, highconf_intervals, args)

    # Create and run comparison pipeline
    comparison_pipeline = _create_comparison_pipeline(vcu, vu, cmp_intervals, highconf_intervals, args)
    raw_calls_vcf, concordance_vcf = comparison_pipeline.run()

    # Process concordance based on interval type
    if not cmp_intervals.is_none():
        # single interval-file concordance - will be saved in a single dataframe
        _process_single_interval_concordance(raw_calls_vcf, concordance_vcf, args)
    else:
        # whole-genome concordance - will be saved in dataframe per chromosome
        _process_whole_genome_concordance(raw_calls_vcf, concordance_vcf, args)


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()
